/*
 *
 *  Copyright 1998  Nothing Real, LLC
 *  All Rights Reserved.
 *
 *  This is UNPUBLISHED PROPRIETARY SOURCE CODE of Nothing Real, LLC.
 *  The contents of this file may not be disclosed to third parties,
 *  copied or duplicated in any form, in whole or in part, without
 *  the prior written permission of Nothing Real, LLC.
 * 
 *  RESTRICTED RIGHTS LEGEND:
 *  Use, duplication or disclosure by the Government is subject to
 *  restrictions as set forth in subdivision (c)(1)(ii) of the Rights
 *  in Technical Data and Computer Software clause at DFARS 252.227-7013,
 *  and/or in similar or successor clauses in the FAR, DOD or NASA FAR
 *  Supplement. Unpublished rights reserved under the Copyright Laws of
 *  the United States.
 *
 *  $Header: /Shake2/scripts/nreal.h 599   6/13/01 11:25a Dan $
 *
 */

/*
 *  #1 DO NOT modify this file under any circumstances.
 *  #2 DO NOT copy this file in its entirety to an include directory
 *  for modification.
 *  #3 DO NOT name a file in your other include directories "nreal.h".
 *  
 *  DO copy the individual items you want to change to an include
 *  directory and customize their values there.
 *
 *  Why!?
 *
 *  #1 This file will change with newer versions of the software. If
 *  you change settings here, they will not apply to a new install.
 *  #2 If (when) this file changes in a new release and you've copied
 *  an old one in its entirety you will very likely override some
 *  items that have been modified for the new relase.
 *  #3 If another file called nreal.h is found, it will be used
 *  instead of this one.  See #2 for why this is bad.
 *
 *  Please do take a look at the section in the HTML manual about
 *  customizing Shake for details on where to put override settings.
 *  
 *  Please do use this file as a reference. It contains lots of
 *  macro definitions, functions available for use in macros and
 *  expressions, convolution kernels, etc. Pretty much anything
 *  non-GUI related that you can customize is shown in this file.
 */


extern "C" {

    // math functions

    int		abs(int);
    double	cbrt(double);
    double	ceil(double);
    double	clamp(double, double, double);
    double	exp(double);
    double	fabs(double);
    double	floor(double);
    double	fmod(double, double);
    double	log(double);
    double	log10(double);
    double	log2(double);
    double	max(double, double);
    double	max3(double, double, double);
    double	min(double, double);
    double	min3(double, double, double);
    double	pow(double, double);
    double	round(double);
    double	sqrt(double);

    float	noise(float);
    float	noise2d(float, float);
    float	noise3d(float, float, float);
    float	noise4d(float, float, float, float);

    float	lnoise(float);
    float	lnoise2d(float, float);
    float	lnoise3d(float, float, float);
    float	lnoise4d(float, float, float, float);

    float	fnoise(float, float);
    float	fnoise2d(float, float, float, float);
    float	fnoise3d(float, float, float, float, float, float);

    float	turbulence(float x, float xScale);
    float	turbulence2d(float x, float y, float xScale, float yScale);
    float	turbulence3d(float x, float y, float z, float xScale, float yScale, float zScale);

    float	rnd(float seed);
    float	rnd1d(float seed, float x);
    float	rnd2d(float seed, float x, float y);
    float	rnd3d(float seed, float x, float y, float z);
    float	rnd4d(float seed, float x, float y, float z, float w);

    // trig functions (work in radians)

    double	acos(double);
    double	asin(double);
    double	atan(double);
    double	atan2(double, double);
    double 	cos(double);
    double 	sin(double);
    double 	tan(double);

    // same as standard trig functions but work in degrees

    double	acosd(double);
    double	asind(double);
    double	atand(double);
    double	atan2d(double, double);
    double 	cosd(double);
    double 	sind(double);
    double 	tand(double);


    double	distance(double x1, double y1, double x2, double y2);

    // misc

    int 	    printf(const char *, ...);
    const char *    stringf(const char *, ...);
    int		    strlen(const char *);
    const char *    strsub(const char *, int index, int len);

    double	atof(const char *);
    int		atoi(const char *);
    long	atol(const char *);

    //

    double	biasedGain(double value, double gain, double bias);

    const char *CRFormatList();
    const char *CWFormatList();
    void	DefCurveType(const char *name, int numArgs, int numControlPoints);

    void	DefRenderNode(const char *ipAddr, int numProcesssors);

    void	DefFormatType(const char *name, int width, int height, float aspectRatio, float viewerAspectRatio, float framesPerSecond, const char* timecodeMode);
    const char	*FormatList();

    void	DefTimecodeMode(const char *name, int fps, int numFramesToDrop=0,int numSecondsDropInterval=0,int numSecondsDropException=0);
    const char	*TimecodeModeList();

    float	Linear(implicit float time, int cycle, const float [2], ....);
    float	LinearV(float, int cycle, const float [2], ....);
    float	LinearLTan(const float [2],....);
    float	LinearRTan(const float [2],....);

    float	CSpline(implicit float time, int cycle, const float [2], ....);
    float	CSplineV(float, int cycle, const float [2], ....);
    float	CSplineLTan(const float [2],....);
    float	CSplineRTan(const float [2],....);

    float	Hermite(implicit float time, int cycle, const float [4], ....);
    float	HermiteV(float, int cycle, const float [4], ....);
    float	HermiteLTan(const float [2],....);
    float	HermiteRTan(const float [2],....);

    float	JSpline(implicit float time, int cycle, const float [2], ....);
    float	JSplineV(float, int cycle, const float [2], ....);
    float	JSplineLTan(const float [2],....);
    float	JSplineRTan(const float [2],....);

    float	NSpline(implicit float time, int cycle, const float [2], ....);
    float	NSplineV(float, int cycle, const float [2], ....);
    float	NSplineLTan(const float [2],....);
    float	NSplineRTan(const float [2],....);

    float	Step(implicit float time, int cycle, const float [2], ....);
    float	StepV(float, int cycle, const float [2], ....);
    float	StepLTan(const float [2],....);
    float	StepRTan(const float [2],....);

    void	SetCacheMode(image, int);
    void	SetFps(curve);
    void	SetFieldRendering(curve);
    void 	SetMotionBlur(curve, curve, curve);
    void	SetPixelScale(curve, curve);
    void	SetProxyScale(curve, curve);
    void	SetProxyFilter(const char *);
    void 	SetQuality(curve);
    void	SetTime(float);
    void	SetTimeRange(const char *);
    void	SetFormat(const char*);
    void	SetVideoResolution(const char*);
    void        SetDefaultWidth(int);
    void        SetDefaultHeight(int);
    void	SetDefaultBytes(int);
    void        SetDefaultAspect(float);
    void        SetDefaultViewerAspect(float);
    void	SetTimecodeMode(const char*);
    int		GetDefaultWidth();
    int		GetDefaultHeight();
    int		GetDefaultBytes();
    float       GetDefaultAspect();
    
    void	LaunchBrowser(const char*,int);
    void	LaunchDocBrowser(const char*);

    image	Crop(image, int left = 0, int bottom = 0, int right = width, int top = height);
    image	Viewport(image, int left = 0, int bottom = 0, int right = width, int top = height);
    image	Window(image, int left = 0, int bottom = 0, int right = width, int top = height);
    image	IRetime(image, 
		    float timeSlide=0, 
		    float inPoint=autoInPoint, 
		    float outPoint=autoOutPoint, 
		    string inMode="Freeze", 
		    string outMode="Freeze"
		);

    const char	*   getInstallPath();
    const char	*   getUserFilesPath();
    const char	*   uncFileName(const char *);

    void	    loadPlugin(const char *);

    // registry

    void	    SetKey(const char *, ....);
    const char *    GetKey(const char *);

    // for internal use

    int	    GetCurveKeys2(int, int, const float[2], ....);
    int	    GetCurveKeys4(int, int, const float[4], ....);
    float   GetVExpr(int,const char*);

    int	    nrcGetAvailableProcessors();

    float   lookupFofx(unsigned int,float, int);
};

extern "libnrfx_n32.so"
{
    // In/Out



    image Black(int width = GetDefaultWidth(), int height = GetDefaultHeight(), int bytes = GetDefaultBytes());
    image Checker(int width = GetDefaultWidth(), int height = GetDefaultHeight(), int bytes = GetDefaultBytes(), int xSize = 32, int ySize = xSize);
    image Color(int width = GetDefaultWidth(), int height = GetDefaultHeight(), int bytes = GetDefaultBytes(), float red = 0, float green = red, float blue = red, float alpha = 1, float depth = 0);
    image ColorWheel(int width = GetDefaultWidth(), int height = GetDefaultHeight(), int bytes = GetDefaultBytes(), float satCenter = 0, float satEdge = 1, float valCenter = 1, float valEdge = 1);
    image Commit(image img, const char *path = "", const char *range = "", ....);
    image FileIn(const char *imageName, const char *fileFormat = "Auto", int autoAlpha = 0, int deInterlacing = 0, ....);
    image FileOut(image, const char *imageName, const char *fileFormat = "Auto", ....);
    image Grad(int width = GetDefaultWidth(), int height = GetDefaultHeight(), int bytes = GetDefaultBytes(), 
	float xCenter = 0.5, float yCenter = 0.5,
	float rLL = 1, float gLL = 0, float bLL = 0, float aLL = 1, float zLL = 0, 
	float rLR = 0, float gLR = 1, float bLR = 0, float aLR = 1, float zLR = 0, 
	float rUR = 0, float gUR = 0, float bUR = 1, float aUR = 1, float zUR = 0, 
	float rUL = 0, float gUL = 0, float bUL = 0, float aUL = 1, float zUL = 0 
    );
    image Julia(int width = GetDefaultWidth(), int height = GetDefaultHeight(), int bytes = GetDefaultBytes(), int iterations=50, float offsetX=0,float ofssetY=0,float zoom=1.0/3.0, float cr = 0.1, float ci = 0.5, float ri=0.2,float gi=0.2,float bi=0.2,float r=1.0,float g=1.0,float b=1.0,....);
    image Mandelbrot(int width = GetDefaultWidth(), int height = GetDefaultHeight(), int bytes = GetDefaultBytes(), int iterations=50, float offsetX=0,float ofssetY=0,float zoom=1.0/3.0,float ri=0.2,float gi=0.2,float bi=0.2,float r=1.0,float g=1.0,float b=1.0,....);
    image QuickShape(int width = GetDefaultWidth(), int height = GetDefaultHeight(), int bytes = GetDefaultBytes(), 
	float xPan = 0 , float yPan = 0, 
	float angle = 0, float aspectRatio = GetDefaultAspect(), 
	float xScale = 1, float yScale = xScale, 
	float xCenter = width/2, float yCenter = height/2, 
	float motionBlur = 0, float shutterTiming = 0.5, float shutterOffset = 0,
	int state = 0,
	....
    );
    image Ramp(int width = GetDefaultWidth(), int height = GetDefaultHeight(), int bytes = GetDefaultBytes(), int orientation = 1, float center = 0.5,
	       float red1 = 0, float green1 = 0, float blue1 = 0, float alpha1 = 1, float depth1 = 0,
	       float red2 = 1, float green2 = 1, float blue2 = 1, float alpha2 = 1, float depth2 = 0);
    image Rand(int width = GetDefaultWidth(), int height = GetDefaultHeight(), int bytes = GetDefaultBytes(), float density = 1, float seed = time);
    image RGrad(int width = GetDefaultWidth() , int height = GetDefaultHeight(), int bytes = GetDefaultBytes(), 
	float xCenter = width/2, float yCenter = height/2,
	float aspectRatio = 1/GetDefaultAspect(),
	float radius = min(width,height)/4, float falloffRadius = min(width,height)/4,
	float fallOff = 0.5,
	float rCenter = 1, float gCenter = 1,float bCenter = 1 ,float aCenter = 1, float zCenter = 0,
	float rEdge = 0, float gEdge = 0,float bEdge = 0,float aEdge = 0, float zEdge = 0
    );
    image Text(int width = GetDefaultWidth(), int height = GetDefaultHeight(), int bytes = GetDefaultBytes(),
	const char *text = "Text", const char *font= "Utopia Regular", 
	float xFontScale = 100, float yFontScale = xFontScale,
	float leading = 1,
	float xPos = width/2, float yPos = height/2, float zPos = 0,
	int xAlign = 2, int yAlign = 2,
	float red = 1, float green = 1, float blue = 1, float alpha = 1,
	float xAngle = 0, float yAngle = 0, float zAngle = 0,
	float fieldOfView = 45,
        float kerning = 0,
        float fontQuality = 1
    );
    image VFileIn(const char *imageName, const char *fileFormat = "Auto", int autoAlpha = 0, int deInterlacing = 0, ....);
    image VFileOut(image, const char *imageName, const char *fileFormat = "Auto",const char *size = "Retain Size", ....);
    image VTRIn( const char* autoAlpha=0, const char* deInterlacing=0, const char* tcm=0, const char* in=0, const char* out=0, const char* duration=0,const char* reel=0, const char* clip=0, ....);
    image VTROut( image, const char* tcm=0, const char* in=0, const char* out=0, const char* duration=0,const char* yuvEncode=0);
    image VTRFileOut( image, const char* tcm, const char* in, const char* out, const char* duration,const char* yuvEncode, const char* frameListFileName, const char* width, const char* height );
    const char *nrfxGetDBFileName(const char *fileName, float time);
    image ZonePlate(int width = GetDefaultWidth(), int height = GetDefaultHeight(), int bytes = GetDefaultBytes(),
	int centerX = width/2, int centerY = height/2, float variance = 1.64, float offset = 0,
	float red1 = 0.0, float green1 = 0.0, float blue1 = 0.0, float alpha1 = 0.0,
	float red2 = 1.0, float green2 = 1.0, float blue2 = 1.0, float alpha2 = 1.0
    );

    // Color

    image Add(image, float red = 0, float green = red, float blue = red, float alpha = 0, float depth = 0);
    image AdjustHSV(image,
	float sourceHue=0, float hueOffset=sourceHue-sourceHue, float hueRange=1, float hueFalloff=0, float hueSharpness=1.5,
	float sourceSat=0, float satOffset=sourceSat-sourceSat, float satRange=1, float satFalloff=0, float satSharpness=1.5,
	float sourceVal=0, float valOffset=sourceVal-sourceVal, float valRange=1, float valFalloff=0, float valSharpness=1.5,
    );
    image Brightness(image, float value = 1);
    image ColorReplace(image in, int affectalpha = 0, 
	float rscolor = 0,float gscolor = 0, float bscolor = 0,
	float rrcolor = rSource,float grcolor = gSource,float brcolor = bSource,
	float low = 0.1,float rolllow = 0.1,
	float high = 0.1,float rollhigh = 0.1,
	float lum = 1,float rolllum = 1);

    image Clamp(image, float rLo = 0, float gLo = rLo, float bLo = rLo, float aLo = rLo, float rHi = 1, float gHi = rHi, float bHi = rHi, float aHi = rHi);
    image CMYToRGB(image);
    image RangeColorCorrect(image,float rExpr,         float gExpr,         float bExpr,         float aExpr,
	float rContrast=1.0, float gContrast=1.0, float bContrast=1.0, float aContrast=1.0,
	float rCenter=0.5,   float gCenter=0.5,   float bCenter=0.5,   float aCenter=0.5,
	float rSoftClip=0.0, float gSoftClip=0.0, float bSoftClip=0.0, float aSoftClip=0.0,
			     float rGamma=1.0,    float gGamma=1.0,    float bGamma=1.0,    float aGamma=1.0,
			     float rGain=1.0,     float gGain=1.0,     float bGain=1.0,     float aGain=1.0,
			     float rAdd=0.0,      float gAdd=0.0,      float bAdd=0.0,      float aAdd=0.0,
			     float rlContrast=1.0,float glContrast=1.0,float blContrast=1.0,float alContrast=1.0,
			     float rlCenter=0.5,  float glCenter=0.5,  float blCenter=0.5,  float alCenter=0.5,
			     float rlSoftClip=0.0,float glSoftClip=0.0,float blSoftClip=0.0,float alSoftClip=0.0,
			     float rlGamma=1.0,   float glGamma=1.0,   float blGamma=1.0,   float alGamma=1.0,
			     float rlGain=1.0,    float glGain=1.0,    float blGain=1.0,    float alGain=1.0,
			     float rlAdd=0.0,     float glAdd=0.0,     float blAdd=0.0,     float alAdd=0.0,
			     float rmContrast=1.0,float gmContrast=1.0,float bmContrast=1.0,float amContrast=1.0,
			     float rmCenter=0.5,  float gmCenter=0.5,  float bmCenter=0.5,  float amCenter=0.5,
			     float rmSoftClip=0.0,float gmSoftClip=0.0,float bmSoftClip=0.0,float amSoftClip=0.0,
			     float rmGamma=1.0,   float gmGamma=1.0,   float bmGamma=1.0,   float amGamma=1.0,
			     float rmGain=1.0,    float gmGain=1.0,    float bmGain=1.0,    float amGain=1.0,
			     float rmAdd=0.0,     float gmAdd=0.0,     float bmAdd=0.0,     float amAdd=0.0,
			     float rhContrast=1.0,float ghContrast=1.0,float bhContrast=1.0,float ahContrast=1.0,
			     float rhCenter=0.5,  float ghCenter=0.5,  float bhCenter=0.5,  float ahCenter=0.5,
			     float rhSoftClip=0.0,float ghSoftClip=0.0,float bhSoftClip=0.0,float ahSoftClip=0.0,
			     float rhGamma=1.0,   float ghGamma=1.0,   float bhGamma=1.0,   float ahGamma=1.0,
			     float rhGain=1.0,    float ghGain=1.0,    float bhGain=1.0,    float ahGain=1.0,
			     float rhAdd=0.0,     float ghAdd=0.0,     float bhAdd=0.0,     float ahAdd=0.0,
			     int rInvert=0,       int   gInvert=0,     int   bInvert=0,     int aInvert=0,
			     float lowWeight=0.5, float highWeight=0.5,const char *tOrder = "GMAC"
			     );
    image ColorCorrect(
	image in, const char* version = "v1.0",
	float rExpr="x",     float gExpr="x",     float bExpr="x",     float aExpr="x",
	float rContrast=1.0, float gContrast=1.0, float bContrast=1.0, float aContrast=1.0,
	float rCenter=0.5,   float gCenter=0.5,   float bCenter=0.5,   float aCenter=0.5,
	float rSoftClip=0.0, float gSoftClip=0.0, float bSoftClip=0.0, float aSoftClip=0.0,
	float rGamma=1.0,    float gGamma=1.0,    float bGamma=1.0,    float aGamma=1.0,
	float rGain=1.0,     float gGain=1.0,     float bGain=1.0,     float aGain=1.0,
	float rAdd=0.0,      float gAdd=0.0,      float bAdd=0.0,      float aAdd=0.0,
	float rlContrast=1.0,float glContrast=1.0,float blContrast=1.0,float alContrast=1.0,
	float rlCenter=0.5,  float glCenter=0.5,  float blCenter=0.5,  float alCenter=0.5,
	float rlSoftClip=0.0,float glSoftClip=0.0,float blSoftClip=0.0,float alSoftClip=0.0,
	float rlGamma=1.0,   float glGamma=1.0,   float blGamma=1.0,   float alGamma=1.0,
	float rlGain=1.0,    float glGain=1.0,    float blGain=1.0,    float alGain=1.0,
	float rlAdd=0.0,     float glAdd=0.0,     float blAdd=0.0,     float alAdd=0.0,
	float rmContrast=1.0,float gmContrast=1.0,float bmContrast=1.0,float amContrast=1.0,
	float rmCenter=0.5,  float gmCenter=0.5,  float bmCenter=0.5,  float amCenter=0.5,
	float rmSoftClip=0.0,float gmSoftClip=0.0,float bmSoftClip=0.0,float amSoftClip=0.0,
	float rmGamma=1.0,   float gmGamma=1.0,   float bmGamma=1.0,   float amGamma=1.0,
	float rmGain=1.0,    float gmGain=1.0,    float bmGain=1.0,    float amGain=1.0,
	float rmAdd=0.0,     float gmAdd=0.0,     float bmAdd=0.0,     float amAdd=0.0,
	float rhContrast=1.0,float ghContrast=1.0,float bhContrast=1.0,float ahContrast=1.0,
	float rhCenter=0.5,  float ghCenter=0.5,  float bhCenter=0.5,  float ahCenter=0.5,
	float rhSoftClip=0.0,float ghSoftClip=0.0,float bhSoftClip=0.0,float ahSoftClip=0.0,
	float rhGamma=1.0,   float ghGamma=1.0,   float bhGamma=1.0,   float ahGamma=1.0,
	float rhGain=1.0,    float ghGain=1.0,    float bhGain=1.0,    float ahGain=1.0,
	float rhAdd=0.0,     float ghAdd=0.0,     float bhAdd=0.0,     float ahAdd=0.0,
	int rInvert=0,       int   gInvert=0,     int   bInvert=0,     int aInvert=0,
	float lowWeight=0.5, float highWeight=0.5,const char *tOrder = "GMAC",
	int affectAlpha = 0,float rSource = 0,float gSource = 0,float bSource = 0,
	float rReplace = "rSource",float gReplace = "gSource",float bReplace = "bSource",
	float low = 0.1,float rolllow = 0.1,float high = 0.1,float rollhigh = 0.1,
	float lum = 1,float rolllum = 1,
	const char *channels = "rgba", int preMult=1,int viewRange=0);

    image ColorX(image, float rExpr = r, float gExpr = g, float bExpr = b, float aExpr = a, float zExpr = z);
    image Compress(image, float rLo = 0, float gLo = rLo, float bLo = rLo, float aLo = rLo, float rHi = 1, float gHi = rHi, float bHi = rHi, float aHi = rHi);
    image ContrastLum(image, float value, float center = 0.5, float smooth = 0);
    image ContrastRGB(image, 
	float rValue, float gValue = rValue, float bValue = rValue, float aValue = rValue, 
	float rCenter, float gCenter = rCenter, float bCenter = rCenter, float aCenter = rCenter, 
	float rSoftClip = 0, float gSoftClip = rSoftClip, float bSoftClip = rSoftClip, float aSoftClip = rSoftClip);
    image DelogC(image, float rOffset = 0, float gOffset = 0, float bOffset = 0, float black = 95, float white = 685, float nGamma = 0.6, float dGamma = 1.0, float softClip = 0);
    image DelogRGB(image, float rOffset = 0, float gOffset = 0, float bOffset = 0, 
	float rBlack = 95, float gBlack = 95, float bBlack = 95, 
	float rWhite = 685, float gWhite = 685, float bWhite = 685, 
	float rNGamma = 0.6, float gNGamma = 0.6, float bNGamma = 0.6, 
	float rDGamma = 1.0, float gDGamma = 1.0, float bDGamma = 1.0, 
	float rSoftClip = 0, float gSoftClip = 0, float bSoftClip = 0);
    image Expand(image, float rLo = 0, float gLo = rLo, float bLo = rLo, float aLo = rLo, float rHi = 1, float gHi = rHi, float bHi = rHi, float aHi = rHi);
    image Fade(image, float value = 1);
    image Gamma(image, float red = 1, float green = rGamma, float blue = rGamma, float alpha = 1);
    image HLSToRGB(image, float rWeight=0.3, float gWeight=0.59, float bWeight=0.11);
    image HSVToRGB(image);
    image Invert(image, const char *channels = "rgba");
    image LogC(image, float rOffset = 0, float gOffset = 0, float bOffset = 0, float black = 95, float white = 685, float nGamma = 0.6, float dGamma = 1.0, float softClip = 0);
    image LogRGB(image, float rOffset = 0, float gOffset = 0, float bOffset = 0, 
	float rBlack = 95, float gBlack = 95, float bBlack = 95, 
	float rWhite = 685, float gWhite = 685, float bWhite = 685, 
	float rNGamma = 0.6, float gNGamma = 0.6, float bNGamma = 0.6, 
	float rDGamma = 1.0, float gDGamma = 1.0, float bDGamma = 1.0, 
	float rSoftClip = 0, float gSoftClip = 0, float bSoftClip = 0);
    image Lookup(image, float rExpr, float gExpr, float bExpr, float aExpr);
    image LookupHLS(image, float hExpr, float lExpr, float sExpr, float aExpr);
    image LookupHSV(image, float hExpr, float sExpr, float vExpr, float aExpr);
    image LookupFile(image, const char *fileName, const char *channels = "rgba");
    image Luminance(image);
    image MDiv(image, int ignoreZero = 0);
    image MMult(image, int ignoreZero = 0);
    image Monochrome(image, float rWeight=0.3, float gWeight=0.59, float bWeight=0.11);
    image Mult(image, float red = 1, float green = 1, float blue = 1, float alpha = 1, float depth = 1);
    image Reorder(image, const char *channels = "rgba");
    image RGBToCMY(image);
    image RGBToHLS(image, float rWeight=0.3, float gWeight=0.59, float bWeight=0.11);
    image RGBToHSV(image);
    image RGBToYIQ(image);
    image RGBToYUV(image);
    image Saturation(image, float value = 1);
    image Set(image, const char *channels="", float value=1);
    image SetBGColor(image, const char *channels="rgbaz", float red = 0, float green = 0, float blue = 0, float alpha = 0, float depth = 0);
    image Solarize(image, float value = 0.2, int thresholdType = 1);
    image Threshold(image=0,float red=0.0,float green=red,float blue=red,float alpha=0.0,int crush=0);
    image VideoSafe(image, int videoType = 0, int processingType = 0, float chromaRange = 100, float compositeRange = 110, float videoGamma = videoType ? 2.8 : 2.2);
    image YIQToRGB(image);
    image YUVToRGB(image);

    // Filter

    image ApplyFilter(image, const char *xFilter = "gauss", const char *yFilter = xFilter, float xScale = 1, float yScale = xScale, int spread = 0, );
    image Blur(image, float xPixels=0, float yPixels = xPixels, int spread = 0, const char *xFilter = "gauss", const char *yFilter = xFilter, const char *channels="rgba");
    image Convolve(image, const char *channels="rgba", const char *kernelName="blur3x3", float percent=100.0, int absolute = 0);
    image Defocus(image, float xPixels=0, float yPixels = xPixels, const char *channels="rgba", float percent=100.0, const char *shape ="circle",
	          float boostPoint = 0.95, float superWhite = 1.0);
    image DilateErode(image, const char *channels, float xPixels, float yPixels = xPixels, int border = 0, int soften = 0, float sharpness = 0);
    image Emboss(image, float gain=1, float azimuth=45, float elevation=30);
    image Grain(image, float size=1, float density=0.8, float seed = time, float filmResponse=-1, float lGain=0.5, float rGain=0.5, float gGain = rGain, float bGain = rGain,
	        float rSize=1.f, float gSize=1.f, float bSize=1.f, float rDensity=1.f, float gDensity=1.f, float bDensity=1.f, float aspect = 1.f,
		float rBias=0.f, float rLow=0.f, float rHigh=0.f,float gBias=0.f, float gLow=0.f, float gHigh=0.f,float bBias=0.f, float bLow=0.f, float bHigh=0.f,
		float softness=0.f, float rSoftness=0.f, float gSoftness=0.f, float bSoftness=0.f, int obeyAlpha=0);
    image Median(image, const char *channels = "rgba", float threshold = 1, int thresholdType = 0);
    image RBlur(image, float xCenter, float yCenter, float iRadius, float oRadius, 
	float aspectRatio = GetDefaultAspect(), float damp=1, float amplitude=0.2, float quality=1, int mirror=0);
    void  DefKernel(const char *kernelName, int xSize, int ySize, float scale, float offset, ....);

    // Keying

    image ChromaKey(image,
	float h, float hRange, float hFRange, float hFSharp,
	float s, float sRange, float sFRange, float sFSharp,
	float v, float vRange, float vFRange, float vFSharp,
	int matteMult = 1, int arith = 0
    );
    image DepthKey(image, float loVal, float hiVal, float loSmooth, float hiSmooth,int matteMult);
    image DepthSlice(image, float center=0, float lo=0, float hi=0, int grad=0, int mirror=0,int matteMult=0);
    image LumaKey(image, float loVal, float hiVal, float loSmooth = 0, float hiSmooth = 0,int matteMult = 0);
    image SpillSuppress(image,float rGain, float gGain, float lumGain , int cLogic);

    // Layer

    image AddText(image img,
	const char *text = "Text", const char *font= "Utopia Regular", 
	float xFontScale = 100, float yFontScale = xFontScale,
	float leading = 1,
	float xPos = width/2, float yPos = height/2, float zPos = 0,
	int xAlign = 2, int yAlign = 2,
	float red = 1, float green = 1, float blue = 1, float alpha = 1,
	float xAngle = 0, float yAngle = 0, float zAngle = 0,
	float fieldOfView = 45,
        float kerning = 0,
        float fontQuality = 1
    );
    image Atop(image, image background, int clipMode = 1);
    image Common(image, image background, int clipMode = 1, int mode=0, float rTol=0, float gTol=0, float bTol=0, float aTol=0);
    image Constraint(image, image background, int cm=1, int type=0, int x1=0, int x2=width, int y1=0, int y2=height, 
	float rTol=0, float gTol=0, float bTol=0, float aTol=0, int thresholdType=0, int tReplace=0, const char *channels="rgba", int field=0, int invert=0);
    image Copy(image, image background, int clipMode = 1, const char *channels="");
    image IAdd(image, image background, int clipMode = 1, int percent = 100);
    image IDiv(image, image background, int clipMode = 1, int percent = 100, int ignoreZero = 1);
    image IMult(image, image background, int clipMode = 1, int percent = 100, int ignoreZero = 0);
    image Inside(image, image background,int clipMode = 1);
    image ISub(image, image background, int clipMode = 1, int percent = 100);
    image ISubA(image, image background, int clipMode = 1, int percent = 100);
    image KeyMix(image, image background, image keyImg, int clipMode = 1, const char *channel = "A", float percent = 100, int invertKey = 0);
    image LayerX(image, image background, float rExpr=r, float gExpr=g, float bExpr=b, float aExpr=a, float zExpr = z);
    image Layer(image, image foreground, const char *operation = "Over",....);
    image Max(image, image background, int clipMode = 1, float percent = 100);
    image Min(image, image background, int clipMode = 1, float percent = 100);
    image Mix(image, image background, int clipMode = 1, float percent = 50, const char *channels = "rgba");
    image Outside(image, image background, int clipMode = 1);
    image Over(image, image background, int clipMode = 1, int preMultiply = 0, int addMattes = 0);
    image Screen(image, image background, int clipMode = 1);
    image SwitchMatte(image img, image matteImg, int clipMode = 1, const char *matteChannel = "A", int matteMult = 1, int invertMatte = 0);
    image Xor(image,image background,int clipMode = 1, int useMatte=0);
    image ZCompose(image, image background, int clipMode = 1, int invert = 0, float tolerance = 0.0);

    // Time

    image TimeX(image, float newTime = time, float newInPoint = inPoint, float newOutPoint = outPoint);

    // Transform

    image CornerPin(image, 
	float x0, float y0, float x1, float y1, float x2, float y2, float x3, float y3,
	const char *xFilter = "default", const char *yFilter = xFilter, 
	int invertTransform = 0,
	float antialiasing = 1,
	float motionBlur = 0, float shutterTiming = 0.5, float shutterOffset = 0,
	int useReference = 0, float referenceFrame = time
    );
    image Fit(image, int xSize=width, int ySize=height, const char *xFilter = "default", const char *yFilter = xFilter, int preCrop = 1);
    image Flip(image);
    image Flop(image);
    image Move2D(image, 
	float xPan = 0 , float yPan = 0, 
	float angle = 0, float aspectRatio = GetDefaultAspect(), 
	float xScale = 1, float yScale = xScale, 
	float xShear = 0, float yShear = 0, 
	float xCenter = width/2, float yCenter = height/2, 
	const char *xFilter = "default", const char *yFilter = xFilter, 
	const char *transformOrder = "trsx", 
	int invertTransform = 0, 
	float motionBlur = 0, float shutterTiming = 0.5, float shutterOffset = 0,
	int useReference = 0, float referenceFrame = time
    );
    image Move3D(image, 
	float xPan = 0, float yPan = 0, float zPan = 0, 
	float xAngle = 0, float yAngle = 0, float zAngle = 0,
	float aspectRatio = GetDefaultAspect(), 
	float xScale = 1, float yScale = xScale, float zScale = yScale,
	float xCenter = width/2, float yCenter = height/2, float zCenter = 0,
	float fov = 0,
	const char *xFilter = "default", const char *yFilter = xFilter, 
	const char *transformOrder = "trs", 
	int invertTransform = 0, 
	float motionBlur = 0, float shutterTiming = 0.5, float shutterOffset = 0,
	int useReference = 0, float referenceFrame = time
    );
    image Move4x4(image, 
	float m00 = 1, float m01 = 0, float m02 = 0, float m03 = 0, 
	float m10 = 0, float m11 = 1, float m12 = 0, float m13 = 0, 
	float m20 = 0, float m21 = 0, float m22 = 1, float m23 = 0, 
	float m30 = 0, float m31 = 0, float m32 = 0, float m33 = 1, 
	const char *xFilter = "default", const char *yFilter = xFilter, 
	int invertTransform = 0, 
	float motionBlur = 0, float shutterTiming = 0.5, float shutterOffset = 0,
	int useReference = 0, float referenceFrame = time
    );
    image Orient(image, int nTurn=0, int flop = 0, int flip = 0);
    image Pan(image, 
	float xPan=0, float yPan=0, 
	float motionBlur = 0, float shutterTiming = 0.5, float shutterOffset = 0
    );
    image Resize(image, float width=width, float height=height, const char *filter = "default", int subPixel = 0);
    image Rotate(image, 
	float angle=0, float aspectRatio = GetDefaultAspect(), 
	float xCenter = width/2, float yCenter = height/2, 
	float motionBlur = 0, float shutterTiming = 0.5, float shutterOffset = 0
    );
    image Scale(image, 
	float xScale=1, float yScale = xScale, 
	float xCenter = width/2, float yCenter = height/2, 
	float motionBlur = 0, float shutterTiming = 0.5, float shutterOffset = 0
    );
//    image SetDOD(image, int left=dod[0], int right=dod[2], int bottom=dod[1], int top=dod[3]);
    image SetDOD(image, int left=0, int right=width, int bottom=0, int top=height);
    image Shear(image, 
	float xShear=0, float yShear = 0, 
	float xCenter = width/2, float yCenter = height/2, 
	float motionBlur = 0, float shutterTiming = 0.5, float shutterOffset = 0
    ); 

    // Transition
    image Append(image, image background);
    image Transition(image i1, image i2, int overlap = 0, const char *mixer = "cut",....);

    // Warp

    image DisplaceX(image, image ctrlImg, int oversamping = 1, float xExpr = x, float yExpr = y, int xDelta = 0, int yDelta = xDelta);
    image IDisplace (image, image ctrlImg, float xScale = 0, float yScale = xScale, float xDOffset = 0, float yDOffset = xDOffset, 
	const char *xChannel = "A", const char *yChannel = "A",
	float xDelta = 0, float yDelta = xDelta);
    image PinCushion(image, int overSampling = 1,float xFactor = 0, float yFactor = 0);
    image Randomize(image, int overSampling = 1, float seed = 0, 
	float xAmplitude = 0, float yAmplitude = xAmplitude,
	float xOffset = 0, float yOffset = 0
    );
    image Turbulate(image, int overSampling = 1, int detail = 1, 
	float xNoiseScale = 10, float yNoiseScale = xNoiseScale, 
	float xAmplitude = 0, float yAmplitude = xAmplitude, 
	float xOffset = 0, 
	float yOffset = 0,
	float seed = 0
    );
    image Twirl(image, 
	float sAngle = 0, float eAngle = 0, 
	float aspectRatio = GetDefaultAspect(),
	float xCenter = width/2, float yCenter = height/2,
	float xRadius = width/2, float yRadius = width/2,
	float bias = 1,
	float antialiasing = 1
    );
    image WarpX(image, int oversamping = 1, float xExpr = x, float yExpr = y, int xDelta = 0, int yDelta = xDelta);
 
    // Other

    image Bytes(image, int outBytes = bytes);
    image Conform(image, int outWidth, int outHeight, int outBytes, string outActive);
    image Field(image, int field);
    image Histogram(image, int width = 512, int height = 512, int maxPerChannel = 1, int ignore = 0);
    image Ignore(image);
    image Info(image);
    image Interlace(image, image fieldImg, int clipMode = 1, int dominance=0, int merge = 0);
    image Mask(image, image mask, const char *maskChannel = "a", float percent = 100, int invertKey = 0, int enableKey = 1);
    image PlotScanline(image, int width = width, int height = 512, int line = 0);
    image QuickPaint(image, image = 0, float = 0, ....);
    image Select(int branch, image img, ....);
    image SwapFields(image);
    image RangeMatch(image, image refImage, int matchHi=1, int matchLo=1, int threshHi = 1, int thresholdLo = 0, 
	             float refFrame=time, int transform=0);
     

    // Viewers

    image View(image, float zoom = 1);
    image Compare(image, float zoom = 1, const char *imageName);
    image Monitor(image, int keepFrames = 0, int width = width, int height = height);

    const char *nfxFilterList();
    const char *nfxKernelList();
    const char *nfxRFormatList();
    const char *nfxWFormatList();
    const char *nfxVFileOutSizeList();

    const char *nfxMixerList();
    void *nfxDefMixer(const char *name, const char *src);
};

extern "libnrpm_n32.so" 
{
    image Primatte (image, image bg=0, image defFg=0, int clipMode = 0, int doWhat = 0, int arith = 0, int doBPix=0, .... );
    image Primatte_v2(image, image bg=0, image gMatte=0, image hMatte=0, image defFg=0, image defBg=0, int version=2, int clipMode=0, 
	int doWhat=0, int arith=0, int doBPix=0, const char *gMatteChannel="a", const char *hMatteChannel="a",
	int useReplacementColor=0, float rR=0.5f, float rG=0.5f, float rB=0.5f, ....);
};

extern "libnrtk_n32.so"
{
    image Tracker(image, const char *trackRange = "1", 
	const char *subPixelRes = "1/16", const char * matchSpace = "luminance",
	float referenceTolerance = 0.75, const char *referenceBehavior = "use start frame",
	float failureTolerance   = 0.5,  const char *failureBehavior   = "stop",
	int limitProcessing = 1, float referenceFrame = 1.0,.... );

    image Stabilize(image,
	int applyTransform = 0, int invertTransform = 0, const char *trackType = "1 point",
	float track1X=width/3, float track1Y=height/3, 
	int stabilizeX = 1, int stabilizeY = 1,
	float track2X=2*width/3, float track2Y=height/3, 
	int scale = 0, int rotation = 0,
	float track3X=2*width/3, float track3Y=2*height/3, 
	float track4X=width/3, float track4Y=2*height/3,
	const char *xFilter = "default", const char *yFilter = xFilter, const char *tOrder="trsx", 
	float motionBlur = 0, float shutterTiming = 0.5, float shutterOffset = 0,
	float referenceFrame = 1,....
    );
    image MatchMove(image, image Foreground,
	int applyTransform = 0, const char *trackType = "1 point",
	float track1X=width/3, float track1Y=height/3, 
	int matchX = 1, int matchY = 1,
	float track2X=2*width/3, float track2Y=height/3, 
	int scale = 1, int rotation = 1,
	float track3X=2*width/3, float track3Y=2*height/3, 
	float track4X=width/3, float track4Y=2*height/3,
	float x1=0, float y1=0,
	float x2=Background.width, float y2=0,
	float x3=Background.width, float y3=Background.height,
	float x4=0, float y4=Background.height,
	const char *xFilter = "default", const char *yFilter = xFilter, 
	float motionBlur = 0, float shutterTiming = 0.5, float shutterOffset = 0,
	float referenceFrame = 1,
	const char *compositeType = "Background", int clipMode=1,const char* version = "v1.1",....
    );

}
const int NGL_POINTS		    = 1;
const int NGL_LINES		    = 2; 
const int NGL_LINE_STRIP	    = 3;
const int NGL_LINE_LOOP		    = 4; 
const int NGL_TRIANGLES 	    = 5;
const int NGL_TRIANGLE_STRIP	    = 6; 
const int NGL_TRIANGLE_FAN	    = 7; 
const int NGL_QUADS		    = 8; 
const int NGL_QUAD_STRIP	    = 9; 
const int NGL_POLYGON		    = 10; 
const int NGL_BEZIER		    = 11; 

const int NGL_ALIGN_XNONE	    = 0;
const int NGL_ALIGN_LEFT	    = 0;
const int NGL_ALIGN_XCENTER	    = 1;
const int NGL_ALIGN_RIGHT	    = 2;
const int NGL_ALIGN_YNONE	    = 0;
const int NGL_ALIGN_BOTTOM	    = 4;
const int NGL_ALIGN_YCENTER	    = 8;
const int NGL_ALIGN_TOP		    = 16;

const int NGL_MODELVIEW		    = 0;
const int NGL_PROJECTION	    = 1;
const int NGL_TEXTURE		    = 2;

const double M_PI = 3.14159265358979323846;

 

extern "libnrgl_n32.so" {

    void	nglFlush();

    void	nglColor3f(float, float, float);
    void	nglColor4f(float, float, float, float);

    void	nglPoint2f(float, float);
    void	nglPointSize(float);

    void	nglMatrixMode(int);
    void	nglPushMatrix();
    void	nglPopMatrix();
    void	nglPerspective(float, float, float, float, float, float);
    void	nglRotatef(float, float, float, float);
    void	nglScalef(float, float, float);
    void	nglTranslatef(float, float, float);

    void	nglBegin(int);
    void	nglEnd();
    void	nglNextContour();

    void	nglVertex2f(float, float);
    void	nglVertex2v(const float *);
    void	nglVertex3f(float, float, float);
    void	nglVertex3v(const float *);

    void	nglFont(const char *, float, float);
    void	nglFontAlignment(int);
    void	nglFontLeading(float);
    void	nglFontKerning(float);
    void	nglFontQuality(float);
    void	nglString(const char *);
    void	nglCMove2f(float, float);

    const char *	nglFontList();

};

// Macros

// In/Out

image Average(const char *imageName, int sStart, int sEnd, int dStart, int dEnd, int bytes, 
		   float gain, float scale)
{
    if(time<dStart || time>dEnd) 
    {
	return TimeX(FileIn(imageName),time<dStart?sStart:sEnd);
    }

    image result;
    float sRange = sEnd-sStart+1;
    float dRange = dEnd-dStart+1;
    float inc = sRange/dRange;
    float sr = (time-dStart)*inc+sStart;
    float er = sr + inc;
    // adjust for scale
    float ssr = sr - (scale-1)*inc/2;
    float ser = er + (scale-1)*inc/2;

    float table[65];
    int samples = sizeof(table)/sizeof(float);
    int maxI = (samples-1)/2;
    table[maxI] = 1.f;
    float sInc = inc*scale/(float)(samples-1);
    float totalWeight = 1;

    float weight = 0;
    if(sRange == dRange && scale <= 1.0)
    {
	result = TimeX(FileIn(imageName),time-(dStart-sStart));
    }
    else
    {
	if(floor(ssr) == floor(ser) || ceil(ssr) == ceil(ser))
	{
	    result = TimeX(FileIn(imageName),floor(ssr));
	}
	else
	{
	    // build table
	    for(int i=0; i<maxI; ++i)
	    {
		table[i] = pow(i/(float)maxI,gain);
		table[samples-i-1] = table[i];
		totalWeight += (2 * table[i]);
	    }
	    for(i=0; i<samples; ++i)
	    {
		table[i] /= totalWeight;
	    }

	    float sframe = floor(ssr);
	    float sframeWeight = 0;
	    int first = 1;
	    float s = ssr;
	    for(i=0; i<=samples; ++i, s+=sInc)
	    {
		float thisFrame = floor(s);
		int v1 = thisFrame>(float)sStart ? 1 : 0;
		int v2 = thisFrame<=(float)sEnd ? 1 : 0;
		int v3 = thisFrame != sframe ? 1 : 0;
		int v4 = (i==samples) && (sframeWeight!=0.0f) ? 1 : 0;
		if(v4 || (v1 && v2 && v3))
		{
		    if(sframe<sStart) { sframe = sStart; }
		    else if(i==samples && thisFrame>sEnd) { sframe = sEnd; }
		    
		    if(first)
		    {
			if(sframeWeight != 0.f)
			{
			    first = 0;
			    result = Bytes(TimeX(FileIn(imageName),sframe),bytes);
			    if(sframeWeight != 1.f) 
			    { 
				SetCacheMode(result, 0);
				result = Fade(result,sframeWeight); 
			    }
			}
		    }
		    else if(sframeWeight != 0.f)
		    {
			SetCacheMode(result, 0);
			result = IAdd(result,Bytes(TimeX(FileIn(imageName),sframe),bytes),0,sframeWeight*100);
		    }
		    sframeWeight = 0;
		    sframe = thisFrame;
		}
		if(i!=samples) { sframeWeight += table[i]; }
	    }
	}
    }
    return result;
}

image PullDown(const char *imageName, int dominance = 0, int offset = 2)
{
    image img1 = TimeX(FileIn(imageName), floor(time*24/30+1));
    image img2 = TimeX(FileIn(imageName), floor(time*24/30));
    image img3 = Interlace(img2, img1, 0, dominance, 0);
    
    return Select(
	1 + (time-1)%5+((time-1<0)?5:0) +4-(offset+1)%5+((offset+1<0)?5:0), 
	img1, img1, img1, img3, img3, img2, img2, img2
    );
}

image PullUp(const char *imageName, int dominance = 0, int offset = 2)
{
    curve lTime = (offset == -1 ? 1 : 0);
    curve lOffset = (offset == -1 ? 3 : offset);
    
    image img1 = TimeX(FileIn(imageName), (floor((time-1)/4)*5+(time-1)%4+1)+lTime);
    image img2 = TimeX(FileIn(imageName), (floor((time-1)/4)*5+(time-1)%4+2)+lTime);
    image img3 = Interlace(img2, img1, 0, dominance, 0);
    
    return Select(
	1 + (time-1)%4+((time-1<0)?5:0) +4-(lOffset+1)%5+((lOffset+1<0)?5:0), 
	img1, img1, img1, img3, img2, img2, img2
	);
}

// Color

image EvenOnly(image img1, image img2)
{
    return Constraint(img2, img1, 0, 8, 0, width, 0, height, 0, 0, 0, 0, 0, 0, "rgba", 0);
}

image Isolate(image img1, image img2, const char * channels)
{
    return Constraint(img2, img1, 0, 4, 0, width, 0, height, 0, 0, 0, 0, 0, 0, channels, 0);
}

image OddOnly(image img1, image img2)
{
    return Constraint(img2, img1, 0, 8, 0, width, 0, height, 0, 0, 0, 0, 0, 0, "rgba", 1);
}

image ROI(image img1, image img2, int left, int bottom, int width, int height)
{
    return Constraint(img2, img1, 0, 1, left, left+::width, bottom, bottom+::height, 0, 0, 0, 0, 0, 0, "rgba", 0);
}

image SetAlpha(image img, float alpha=1)
{
    return Crop(Set(img,"a",alpha), 0,0,width,height);
}

// Filter

image PercentBlur(image img, float xPercent = 0, float yPercent = xPercent, int spread = 0, const char *channels="rgba")
{
    return Blur(img, width*xPercent/100, height*yPercent/100, spread, "gauss", "gauss", channels);
}

image Sharpen(image img, float percent, float xPixels, float yPixels = xPixels, const char *channels="rgba")
{
    return Mix(img, 
	Blur(img, 
	    xPixels, yPixels, 
	    0, 
	    "gauss", "gauss", 
	    channels
	),
	0, 
	percent > -100 ? -percent : 100, 
	"rgbaz"
    );
}

// Layer

image Under(image Background, image Foreground, int clipMode = 1, int preMultiply = 0, int addMattes = 0)
{
    return Over(Foreground, Background, clipMode?0:1, preMultiply, addMattes);
}

// Time

image Cycle(image img, float first, float last, int mirror)
{
    return TimeX(img,
	{{
	    float   t = (time-first);
	    float   length = last-first+1;
	    float   l = (length <= 1 ? 1 : mirror ? length-1 : length);
	    int	    d = (t<0 ? 1-t/l : t/l);
	    float   r = (t<0 ? l-(-t)%l : t%l);
	    mirror ? first+(d%2? l-r : r) : first+r;
	}}
    );
}

image Freeze(image img, float first, float last)
{
    return TimeX(img, time<first ? first : time > last ? last : time);
}

image Repeat(image img, int repetitions)
{
    return TimeX(img, 1+floor((time-1)/(repetitions==0 ? 1 : repetitions < 0 ? -repetitions : repetitions)));
}

image Reverse(image img, float first, float last)
{
    return TimeX(img, last-(time-first));
}

image Sequence2(image img1, image img2, float cut)
{
     return Select(time<cut?1:2,img1, img2);
}

image Sequence3(image img1, image img2, image img3, float cut1, float cut2)
{
     return Select(time<cut1?1:time<cut2?2:3,img1, img2, img3);
}

image TimeShift(image img, float oldStart, float newStart)
{
    curve tOff = newStart - oldStart; 
    return TimeX(img, time - tOff, inPoint + tOff, outPoint + tOff );
}

// Transform

image Scroll(image img, float xScroll=0, float yScroll=0, float motionBlur=0, float shutterTiming=0.5, float shutterOffset=0)
{
    curve lXScroll = (xScroll < 0 ? -(-xScroll)%img.width : xScroll%img.width);
    curve lYScroll = (yScroll < 0 ? -(-yScroll)%img.height : yScroll%img.height);

    Crop1 = Crop(img, 0, 0, width, height);
    Pan1 = Pan(Crop1, lXScroll, lYScroll, motionBlur, shutterTiming, shutterOffset);
    Window1 = Window(Pan1, 0, 0, width, height);
    Window2 = Window(Pan1, lXScroll >= 0 ? width : -width, 0, width, height);
    Window3 = Window(Pan1, 0, lYScroll >= 0 ? height : -height, width, height);
    Window4 = Window(Pan1, lXScroll >= 0 ? width : -width, lYScroll >= 0 ? height : -height, width, height);
    IAdd1 = IAdd(Window1, Window2, 1, 100);
    IAdd2 = IAdd(IAdd1, Window3, 1, 100);
    IAdd3 = IAdd(IAdd2, Window4, 1, 100);

    return IAdd3;
}

image Zoom(image img, float xScale, float yScale = xScale, const char *filter = "default",int subPixel=0)
{
    return Resize(img,width*xScale,height*yScale,filter,subPixel);
}


// Other

image AddBorders(image img, int xBorder = 0, int yBorder = 0)
{
    return Viewport(img, -xBorder, -yBorder, width+xBorder, height+yBorder);
}

image AddShadow(image img, float xOffset = 0, float yOffset = 0, float fuzziness = 0, 
    float rShadow = 0, float gShadow = 0, float bShadow = 0, float opacity = 1)
{
    image shadow =  Pan(
			Mult(
			    Reorder(
				Blur(img, fuzziness, fuzziness, 1, "gauss", "gauss"),
				"aaaa"
			    ),
			    rShadow*opacity,
			    gShadow*opacity,
			    bShadow*opacity,
			    opacity,
			    1
			),
			xOffset,
			yOffset,
			0,0,0
    );
    return Over(img, shadow, 0);
}

image CameraShake(image img, float xFrequency = 1, float yFrequency = xFrequency, 
		  float xAmplitude = 10, float yAmplitude = xAmplitude,
		  float seed = 0, 
		  float motionBlur = 0, float shutterTiming = 0.5, float shutterOffset = 0)
{
    return Move2D(
	img,
	(noise3d(time*xFrequency, seed, 0)*2-1)*xAmplitude,
	(noise3d(time*yFrequency, seed, 100)*2-1)*yAmplitude,
	0,
	1,
	1,
	1,
	0,
	0,
	0,
	0,
	"default",
	"default",
	"rts",
	0,
	motionBlur,
	shutterTiming,
	shutterOffset,
	0,
	1
    );
}

image Deinterlace(image img, int field, int mode = 0)
{
    
    // interp
    p1 = Pan(img, 0, 1, 0, 0.5, 0);
    p2 = Pan(img, 0, -1, 0, 0.5, 0);
    m1 = Mix(p2, p1, 0, 50, "rgbaz");
    pinterp = Interlace(img, m1, 0, ::field, 0);
    
    // blur
    pblur =  Mix(m1, img, 0, 50, "rgbaz");
    
    // replace
    f1 = Field(img,::field);
    preplace = Interlace(f1,f1,0,::field,1);
    
    return Select(mode==0?1:mode==1?2:3, preplace, pinterp, pblur );
}

image DropShadow(image img, float fuzziness = 0, float rShadow = 0, float gShadow = 0, float bShadow = 0, float opacity = 1)
{
    return Mult(
	Reorder(
	    Blur(img, fuzziness, fuzziness, 1, "gauss", "gauss"),
	    "aaaa"
	),
	rShadow*opacity,
	gShadow*opacity,
	bShadow*opacity,
	opacity,
	1
    );
}

image Pixelize(image img, float xPixels, float yPixels = xPixels)
{
    return Resize(
	Resize(img, 
	    (width/script.proxyScale)/(xPixels<1?1:xPixels), 
	    (height/script.proxyScale)/(yPixels<1?1:yPixels), 
	    "gauss", 
	    0
	), 
	img.width, 
	img.height, 
	"dirac", 
	0
    );
}

image Tile(image img, int nXTile=1, int nYTile=1)
{
    image oneTile = Crop(Resize(Crop(img, 0,0,width,height), 
			    (float)width/nXTile, (float)height/nYTile, "default", 1),
			    0, 0, img.width, height);
    image oneRow = oneTile;
    for (int i=1; i<nXTile; ++i) {
	oneRow = Screen(Pan(oneTile, width/nXTile*i,0,0,0,0), oneRow, 0);
	if (nYTile > 1 || i < nXTile-1) {
	    SetCacheMode(oneRow, 0);
	}
    }
    oneTile = Crop(oneRow, 0,0,img.width,img.height);
    oneRow = oneTile;
    for (int j=1; j<nYTile; ++j) {
	oneRow = Screen(Pan(oneTile, 0,height/nYTile*j,0,0,0), oneRow, 0);
	if (j < nYTile-1) {
	    SetCacheMode(oneRow, 0);
	}
    }
    return oneRow;
}

image Stars(image img, float radius = width/4, float scale = 0.4, int nBranch = 3, int levels = 1)
{
    // simple example of recursive function

    image result = img;
    if (levels > 1) {
	result = Stars(img, radius, scale, nBranch, levels-1);
    }
    if (nBranch >= 1) {
	image oneBranch = Pan(Scale(result, scale), radius, 0);
	
	for (int i=0; i<nBranch; ++i) {
	    result = Over(Rotate(oneBranch, i*360.0/nBranch), result);
	}
    }
    return result;
}

image ZBlur(image img, float amount=20.0, float near = 0.0, float far = 1.0, float focusCenter=(far-near)/2+near, float focusRange=0.0, int steps=5, float stepBlend=1.0)
{    
    image result = img;
    curve float inFocusRange = focusRange/2.0;
    // find the number of regions and step params
    curve float regionStep = (near-(focusCenter-(focusRange/2.0)))/steps;
    curve float begin = focusCenter-inFocusRange;
    curve float regionStep2 = (far-(focusCenter+(focusRange/2.0)))/steps;
    curve float begin2 = focusCenter+inFocusRange;
    curve float blurStep = amount/steps;
    for (int i=0; i<steps; ++i) {
	blur1 = Blur(img,blurStep*(i+1),blurStep*(i+1),0,"gauss","gauss");
	// Do the blur before the focus plane
	DepthSlice1 = DepthSlice(img,0,(begin+(i*regionStep)),(begin+(i*regionStep)+(stepBlend*regionStep)),1);					
	result = KeyMix(result,blur1,DepthSlice1);
	DepthSlice1 = DepthSlice(img,0,(begin+(i*regionStep)+(stepBlend*regionStep)),(begin+((i+1)*regionStep)+(stepBlend*regionStep)),0);
	result = KeyMix(result,blur1,DepthSlice1);
	// Do the blur after the focus plane
	DepthSlice1 = DepthSlice(img,0,(begin2+(i*regionStep2)),(begin2+(i*regionStep2)+(stepBlend*regionStep2)),1);					
	result = KeyMix(result,blur1,DepthSlice1);
	DepthSlice1 = DepthSlice(img,0,(begin2+(i*regionStep2)+(stepBlend*regionStep2)),(begin2+((i+1)*regionStep2)+(stepBlend*regionStep2)),0);
	result = KeyMix(result,blur1,DepthSlice1);
    }
    // Blur everything before the near and past the far
    blur1 = Blur(img,amount,amount,0,"gauss","gauss");
    DepthSlice1 = DepthSlice(img,0,near,far,0);					
    result = KeyMix(result,blur1,DepthSlice1, 0,"a",100,1);    
    return Select(((amount > 0) && (near != far) && (focusRange >= 0.0f))?2:1,img,result);	
}

image ZDefocus(image img ,float xPixels=0, float yPixels=xPixels, const char* channels = "rgba", 
	       float percent=100, const char *shape ="circle", float boostPoint=0.95, float superWhite=1,
               float near = 0.0, float far = 1.0, float focusCenter=(far-near)/2+near, float focusRange=0.0, int steps=5, float stepBlend=1.0)
{    
    image result = img;
    curve float inFocusRange = focusRange/2.0;
    // find the number of regions and step params
    curve float regionStep = (near-(focusCenter-(focusRange/2.0)))/steps;
    curve float begin = focusCenter-inFocusRange;
    curve float regionStep2 = (far-(focusCenter+(focusRange/2.0)))/steps;
    curve float begin2 = focusCenter+inFocusRange;
    curve float dfXStep = yPixels/steps;
    curve float dfYStep = xPixels/steps;
    for (int i=0; i<steps; ++i) {
	df1 = Defocus(img,dfXStep*(i+1),dfYStep*(i+1),channels,percent,shape,boostPoint,superWhite);
	// Do the blur before the focus plane
	DepthSlice1 = DepthSlice(img,0,(begin+(i*regionStep)),(begin+(i*regionStep)+(stepBlend*regionStep)),1);					
	result = KeyMix(result,df1,DepthSlice1);
	DepthSlice1 = DepthSlice(img,0,(begin+(i*regionStep)+(stepBlend*regionStep)),(begin+((i+1)*regionStep)+(stepBlend*regionStep)),0);
	result = KeyMix(result,df1,DepthSlice1);
	// Do the blur after the focus plane
	DepthSlice1 = DepthSlice(img,0,(begin2+(i*regionStep2)),(begin2+(i*regionStep2)+(stepBlend*regionStep2)),1);					
	result = KeyMix(result,df1,DepthSlice1);
	DepthSlice1 = DepthSlice(img,0,(begin2+(i*regionStep2)+(stepBlend*regionStep2)),(begin2+((i+1)*regionStep2)+(stepBlend*regionStep2)),0);
	result = KeyMix(result,df1,DepthSlice1);
    }
    // Blur everything before the near and past the far
    blur1 = Blur(img,amount,amount,0,"gauss","gauss");
    DepthSlice1 = DepthSlice(img,0,near,far,0);					
    result = KeyMix(result,df1,DepthSlice1, 0,"a",100,1);    
    return Select(((xPixels > 0) && (yPixels > 0) && (near != far) && (focusRange >= 0.0f))?2:1,img,result);	
}

image IBlur(image img, image controlImg ,float xPixels, float yPixels = xPixels, int spread = 0, 
			const char *xFilter = "gauss", const char *yFilter = xFilter, int steps = 5,
			float stepBlend=1.0, const char* controlChannel ="A", const char* channels = "rgba", int invert=0)
{    
    image result = img;
    image zImg = Reorder(
	controlImg,
	{{
	    const char *mc = controlChannel;
	    (mc[0] == 'r') || (mc[0] == 'R') ? "0000r" : 
	    (mc[0] == 'g') || (mc[0] == 'G') ? "0000g" : 
	    (mc[0] == 'b') || (mc[0] == 'B') ? "0000b" : 
	    (mc[0] == 'a') || (mc[0] == 'A') ? "0000a" : "0000z"; 
	}}
    );
    
    // find the number of regions and step params
    curve float regionStep = (invert ? -1.0f/steps : 1.0f/steps);
    curve float begin = (invert ? 1 : 0);
    curve float blurStepX = xPixels/steps;
    curve float blurStepY = yPixels/steps;
    for (int i=0; i<steps; ++i) {
	// Create the blurred image
	blur1 = Blur(img,blurStepX*(i+1),blurStepY*(i+1),spread,xFilter,yFilter,channels);
	// Do the blur before the focus plane
	DepthSlice1 = DepthSlice(zImg,0,begin+(i*regionStep),begin+(i*regionStep)+(stepBlend*regionStep),1);					
	result = KeyMix(result,blur1,DepthSlice1);
	DepthSlice1 = DepthSlice(zImg,0,begin+(i*regionStep)+(stepBlend*regionStep),begin+((i+1)*regionStep)+(stepBlend*regionStep),0);
	result = KeyMix(result,blur1,DepthSlice1);
    }
    return Select( ((xPixels == 0.f) && (yPixels == 0.f))?1:2,img,result);	
}

image ISharpen(image img, image controlImg ,float percent, float xPixels, float yPixels = xPixels,
	        int steps = 5,float stepBlend=1.0, const char* controlChannel ="A", const char* channels = "rgba", int invert=0)
{    
    image result = img;
    image zImg = Reorder(
	controlImg,
	{{
	    const char *mc = controlChannel;
	    (mc[0] == 'r') || (mc[0] == 'R') ? "0000r" : 
	    (mc[0] == 'g') || (mc[0] == 'G') ? "0000g" : 
	    (mc[0] == 'b') || (mc[0] == 'B') ? "0000b" : 
	    (mc[0] == 'a') || (mc[0] == 'A') ? "0000a" : "0000z"; 
	}}
    );
    
    // find the number of regions and step params
    curve float regionStep = (invert ? -1.0f/steps : 1.0f/steps);
    curve float begin = (invert ? 1 : 0);
    curve float sharpenStepX = xPixels/steps;
    curve float sharpenStepY = yPixels/steps;
    for (int i=0; i<steps; ++i) {
	// Create the blurred image
	sharpen1 = Sharpen(img,percent, sharpenStepX*(i+1),sharpenStepY*(i+1),channels);
	// Do the blur before the focus plane
	DepthSlice1 = DepthSlice(zImg,0,(begin+(i*regionStep)),(begin+(i*regionStep)+(stepBlend*regionStep)),1,0,0);					
	result = KeyMix(result,sharpen1,DepthSlice1);
	DepthSlice1 = DepthSlice(zImg,0,(begin+(i*regionStep)+(stepBlend*regionStep)),(begin+((i+1)*regionStep)+(stepBlend*regionStep)),0,0,0);
	result = KeyMix(result,sharpen1,DepthSlice1);
    }
    return Select( ((xPixels == 0.f) && (yPixels == 0.f))?1:2,img,result);	
}

image IDilateErode(image img, image controlImg ,const char* channels = "rgba", float xPixels, float yPixels = xPixels,
		   int border = 0, int soften = 0, float sharpness = 0, int steps = 5,float stepBlend=1.0, const char* controlChannel ="a",int invert=0)
{    
    image result = img;
    image zImg = Reorder(
	controlImg,
	{{
	    const char *mc = controlChannel;
	    (mc[0] == 'r') || (mc[0] == 'R') ? "0000r" : 
	    (mc[0] == 'g') || (mc[0] == 'G') ? "0000g" : 
	    (mc[0] == 'b') || (mc[0] == 'B') ? "0000b" : 
	    (mc[0] == 'a') || (mc[0] == 'A') ? "0000a" : "0000z"; 
	}}
    );
    
    // find the number of regions and step params
    curve float regionStep = (invert ? -1.0f/steps : 1.0f/steps);
    curve float begin = (invert ? 1 : 0);
    curve float DEStepX = xPixels/steps;
    curve float DEStepY = yPixels/steps;
    for (int i=0; i<steps; ++i) {
	// Create the blurred image
	DE1 = DilateErode(img,channels,DEStepX*(i+1),DEStepY*(i+1),border,soften,sharpness);
	// Do the blur before the focus plane
	DepthSlice1 = DepthSlice(zImg,0,begin+(i*regionStep),begin+(i*regionStep)+(stepBlend*regionStep),1);					
	result = KeyMix(result,DE1,DepthSlice1);
	DepthSlice1 = DepthSlice(zImg,0,begin+(i*regionStep)+(stepBlend*regionStep),begin+((i+1)*regionStep)+(stepBlend*regionStep),0);
	result = KeyMix(result,DE1,DepthSlice1);
    }
    return Select( ((xPixels == 0.f) && (yPixels == 0.f))?1:2,img,result);	
}
image IDefocus(image img, image controlImg ,float xPixels, float yPixels=xPixels, const char* channels = "rgba", 
	       float percent=100, const char *shape ="circle", float boostPoint=0.95, float superWhite=1,
	       int steps = 5,float stepBlend=1.0, const char* controlChannel ="a",int invert=0)
{    
    image result = img;
    image zImg = Reorder(
	controlImg,
	{{
	    const char *mc = controlChannel;
	    (mc[0] == 'r') || (mc[0] == 'R') ? "0000r" : 
	    (mc[0] == 'g') || (mc[0] == 'G') ? "0000g" : 
	    (mc[0] == 'b') || (mc[0] == 'B') ? "0000b" : 
	    (mc[0] == 'a') || (mc[0] == 'A') ? "0000a" : "0000z"; 
	}}
    );
    
    // find the number of regions and step params
    curve float regionStep = (invert ? -1.0f/steps : 1.0f/steps);
    curve float begin = (invert ? 1 : 0);
    curve float xStep = xPixels/steps;
    curve float yStep = yPixels/steps;
    for (int i=0; i<steps; ++i) {
	// Create the blurred image
	df1 = Defocus(img,xStep*(i+1),yStep*(i+1),channels,percent,shape,boostPoint,superWhite);
	// Do the blur before the focus plane
	DepthSlice1 = DepthSlice(zImg,0,begin+(i*regionStep),begin+(i*regionStep)+(stepBlend*regionStep),1);					
	result = KeyMix(result,df1,DepthSlice1);
	DepthSlice1 = DepthSlice(zImg,0,begin+(i*regionStep)+(stepBlend*regionStep),begin+((i+1)*regionStep)+(stepBlend*regionStep),0);
	result = KeyMix(result,df1,DepthSlice1);
    }
    return Select(((xPixels == 0.f) && (yPixels == 0.f))?1:2,img,result);	
}

//IRBlur by Peter Warner 6/24

image IRBlur(

	image img, 
	image controlImg ,
	float xCenter,
	float yCenter,
	float iRadius,
	float oRadius,
	float aspectRatio,
	float damp,
	float amplitude, 
	float blurQuality,
	float mirror,
	int steps = 5,
	float stepBlend=1.0, 
	const char* controlChannel ="a", 
	int invert=0
	)
{    
    image result = img;

    // Move the user selected channel to Z for DepthSlice

    image zImg = Reorder(
	controlImg,
	{{
	    const char *mc = controlChannel;
	    stringf("0000%c", mc[0]);
	}}
    );
    
    // find the number of regions and step params
    curve float regionStep = (invert ? -1.0f/steps : 1.0f/steps);
    curve float begin = (invert ? 1.0f : 0.0f);
    curve float blurStep = amplitude/steps;
    for (int i=0; i<steps; ++i) {
	// Create the blurred image
	blur1 = RBlur(img, xCenter, yCenter, iRadius, oRadius, aspectRatio, damp, blurStep*(i+1), blurQuality, mirror);
	// Do the blur before the focus plane
	DepthSlice1 = DepthSlice(zImg,0,begin+(i*regionStep),begin+(i*regionStep)+(stepBlend*regionStep),1);					
	result = KeyMix(result,blur1,DepthSlice1);
	DepthSlice1 = DepthSlice(zImg,0,begin+(i*regionStep)+(stepBlend*regionStep),begin+((i+1)*regionStep)+(stepBlend*regionStep),0);
	result = KeyMix(result,blur1,DepthSlice1);
    }
    return Select( (amplitude == 0.f) ?1:2,img,result);	
}

image Overlays(image img, int ChartOverlay = 0, int VideoSafe = 1, int MatteOverlay = 0)
{
    
    OneEightFive = Text(img.width, img.height, 1, "1.85", "Courier", round(img.height*.05), xFontScale, 1, round(img.width*.01), round(img.height*.01), 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 45);
    TITLE_SAFE = Text(img.width, img.height, 1, "Title Safe", "Courier", round(Over2.height*.05), xFontScale, 1, round((img.width*.205)/2+4), round((img.height*.2)/2+4), 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 45);
    TV_SAFE = Text(img.width, img.height, 1, "TV Safe", "Courier", round(Over2.height*.05), xFontScale, 1, round((img.width*.1)/2+4), round((img.height*.1)/2+4), 0, 1, 1, 1, 1, 1, 1, 0, 0, 0, 45);
    TV_Safe = Color(img.width*.9, img.height*.9, 1, 1, 1, 1, 1, 0);
    Title_Safe = Color(img.width*.795, img.height*.8, 1, 1, 1, 1, 1, 0);
    fullframe = Color(img.width, img.height, 1, 1, 1, 1, 1, 0);
    horiz = Color(round(img.width/14.4), 3, 1, 1, red, red, 1, 0);
    vert = Color(3, round(img.width/14.4), 1, 1, red, red, 1, 0);
    Crop1 = Crop(Title_Safe, -3, -3, width-6, height-6);
    Crop2 = Crop(TV_Safe, -3, -3, width-6, height-6);
    Crop4 = Crop(fullframe, 0, 0, width, width/1.85);
    Crop5 = Crop(fullframe, -3, -3, width-6, width/1.85-6);
    Pan1 = Pan(TV_Safe, round((img.width*.1)/2), round((img.height*.1)/2), 0, 0.5, 0);
    Pan3 = Pan(horiz, (img.width-horiz.width)/2, (img.height-horiz.height)/2, 0, 0.5, 0);
    Pan4 = Pan(Title_Safe, round((img.width*.205)/2), round((img.height*.2)/2), 0, 0.5, 0);
    Pan6 = Pan(vert, (img.width-vert.width)/2, (img.height-vert.height)/2, 0, 0.5, 0);
    VideoSafe1 = VideoSafe(img, 0, 0, 100, 110, videoType ? 2.2 : 2.8);
    Common1 = Common(VideoSafe1, img, 1, 0, 0, 0, 0, 0);
    Outside3 = Outside(Crop4, Crop5, 0);
    Over1 = Over(Pan3, Pan6, 1, 0, 0);
    Pan2 = Pan(Crop1, round((img.width*.205)/2), round((img.height*.2)/2), 0, 0.5, 0);
    Pan5 = Pan(Crop2, round((img.width*.1)/2), round((img.height*.1)/2), 0, 0.5, 0);
    Outside1 = Outside(Pan1, Pan5, 0);
    Outside2 = Outside(Pan4, Pan2, 0);
    Over7 = Over(OneEightFive, Outside3, 1, 0, 0);
    VideoSafeSelect = Select(VideoSafe+1, img, Common1,  0, 0);
    Add1 = Add(VideoSafeSelect, 1, 0, 0, 0, 0);
    Over2 = Over(Outside1, Outside2, 1, 0, 0);
    Pan12 = Pan(Over7, 0, round((img.height-Crop5.height)/2), 0, 0.5, 0);
    Mask1 = Mask(Add1, VideoSafeSelect, "A", 100, 0, 1);
    MatteOverlaySelect = Select(MatteOverlay+1, VideoSafeSelect, Add1, 0, 0);
    Over3 = Over(TV_SAFE, Over2, 1, 0, 0);
    Over6 = Over(Over1, Pan12, 1, 0, 0);
    Over4 = Over(TITLE_SAFE, Over3, 1, 0, 0);
    Over5 = Over(Over1, Over4, 1, 0, 0);
    ChartOverlaySelect = Select(ChartOverlay+1, Over5, Over6, 0, 0);
    return ISubA(ChartOverlaySelect, MatteOverlaySelect, 1, 100);
}

image RePin(image img,
		float x0Source=0, float y0Source=0, float x1Source=width, float y1Source=0, float x2Source=width, float y2Source=height, float x3Source=0, float y3Source=height,
		float x0=0,   float y0=0,   float x1=width,   float y1=0,   float x2=width,   float y2=height,   float x3=0,   float y3=height,
		const char *xFilter = "default", const char *yFilter = xFilter, 
		int inverseTransform = 0,
		float antialiasing = 1,
		float motionBlur = 0, float shutterTiming = 0.5, float shutterOffset = 0,
		int useReference = 0, float referenceFrame = time)
{
//    CornerPin1 = CornerPin(img, x0Source, y0Source, x1Source, y1Source, x2Source, y2Source, x3Source, y3Source, xFilter, yFilter, 1, antialiasing, motionBlur, shutterTiming, shutterOffset, useReference, referenceFrame);
    return CornerPin(img, x0, y0, x1, y1, x2, y2, x3, y3, xFilter, yFilter, inverseTransform, antialiasing, motionBlur, shutterTiming, shutterOffset, useReference, referenceFrame);
}

//Layer Modified 12/9/99

image LayerMacro(
image	Background, 
image	Foreground, 
int	clipMode= 1,
string	operation= "Over",
float	fgOpacity= 100,
float	bgPercent= 100,
string	channels= "rgba", 
int	preMultiply= 0, 
int	addMattes= 0, 
int	useMatte= 0,
int	ignoreZero= 0
)
{
    image result = Foreground;

    MMultFG = MMult(result, 0);
    FGOpacityMM = Fade(MMultFG,fgOpacity/100);

    FGOpacity = Fade(result,fgOpacity/100);
    
    SelectFGPreMult = Select(preMultiply+1, FGOpacity, FGOpacityMM, 0, 0);
    BGOpacity = Fade(Background,bgPercent/100);

    if (operation == "Atop") {
		result = Atop(SelectFGPreMult, BGOpacity, clipMode);
    } 
    else if (operation == "IAdd") {
	result = IAdd(SelectFGPreMult, Background, clipMode, bgPercent);
    }
    else if (operation == "IDiv") {
	result = IDiv(SelectFGPreMult, Background, clipMode, bgPercent, ignoreZero);
    }
    else if (operation == "IMult") {
	result = IMult(SelectFGPreMult, Background, clipMode, bgPercent, ignoreZero);
    }
    else if (operation == "Inside") {
	result = Inside(SelectFGPreMult, BGOpacity, clipMode);
    } 
    else if (operation == "ISub") {
	result = ISub(SelectFGPreMult, Background, clipMode, bgPercent);
    }
    else if (operation == "ISubA") {
	result = ISubA(SelectFGPreMult, Background, clipMode, bgPercent);
    }
    else if (operation == "Max") {
	result = Max(SelectFGPreMult, Background, clipMode, bgPercent);
    }
    else if (operation == "Min") {
	result = Min(SelectFGPreMult, Background, clipMode, bgPercent);
    }
    else if (operation == "Mix") {
	result = Mix(SelectFGPreMult, Background, clipMode, bgPercent, "rgbaz");
    } 
    else if (operation == "Outside") {
	result = Outside(SelectFGPreMult, BGOpacity, clipMode);
    } 
    else if (operation == "Over") {
	result = Over(SelectFGPreMult, BGOpacity, clipMode, 0, addMattes);
    } 
    else if (operation == "Screen") {
	result = Screen(SelectFGPreMult, BGOpacity, clipMode);
    } 
    else if (operation == "Under") {
	result = Under(SelectFGPreMult, BGOpacity, clipMode, 0, addMattes);
    } 
    else if (operation == "Xor") {
	result = Xor(SelectFGPreMult, BGOpacity, clipMode, useMatte);
    } 
    else if (operation == "Background") {
	result = Background;
    }
    else if (operation == "Modified Background") {
	result = BGOpacity;
    }
    else if (operation == "Modified Foreground") {
	result = SelectFGPreMult;
    }
    return Copy(Foreground, result, 1, channels);	    
}

// prw 10/1/00

image ColorSpace(
	image input, 
	string inSpace = "rgb", 
	string outSpace = "rgb", 
	float rWeight = .3,
	float gWeight = .59,
	float bWeight = .11
)
{
    image result = input;

//CMY in
    if (inSpace == "cmy") && (outSpace == "cmy") {
	result = input;
    }
    else if (inSpace == "cmy") && (outSpace == "hls" ){
	result1 = CMYToRGB(input);
	result = RGBToHLS(result1);
    }	
    else if (inSpace == "cmy") && (outSpace == "hsv" ){
	result1 = CMYToRGB(input);
	result = RGBToHSV(result1);
    }	
    else if (inSpace == "cmy") && (outSpace == "rgb" ){
	result = CMYToRGB(input);
    }	
    else if (inSpace == "cmy") && (outSpace == "yiq" ){
	result1 = CMYToRGB(input);
	result = RGBToYIQ(result1);
    }	
    else if (inSpace == "cmy") && (outSpace == "yuv" ){
	result1 = CMYToRGB(input);
	result = RGBToYUV(result1);
    }	

// HLS in
    else if (inSpace == "hls") && (outSpace == "cmy" ){
	result1 = HLSToRGB(input);
	result = RGBToCMY(result1);
    }	
    else if (inSpace == "hls") && (outSpace == "hls" ){
	result = input;
    }
    else if (inSpace == "hls") && (outSpace == "hsv" ){
	result1 = HLSToRGB(input);
	result = RGBToHSV(result1);
    }	
    else if (inSpace == "hls") && (outSpace == "rgb" ){
	result = HLSToRGB(input,rWeight,gWeight,bWeight);
    }	
    else if (inSpace == "hls") && (outSpace == "yiq" ){
	result1 = HLSToRGB(input);
	result = RGBToYIQ(result1);
    }	
    else if (inSpace == "hls") && (outSpace == "yuv" ){
	result1 = HLSToRGB(input);
	result = RGBToYUV(result1);
    }	

// HSV in
    else if (inSpace == "hsv") && (outSpace == "cmy" ){
	result1 = HSVToRGB(input);
	result = RGBToCMY(result1);
    }	
    else if (inSpace == "hsv") && (outSpace == "hls" ){
	result1 = HSVToRGB(input);
	result = RGBToHLS(result1);
    }	
    else if (inSpace == "hsv") && (outSpace == "hsv" ){
	result = input;
    }
    else if (inSpace == "hsv") && (outSpace == "rgb" ){
	result = HSVToRGB(input);
    }	
    else if (inSpace == "hsv") && (outSpace == "yiq" ){
	result1 = HSVToRGB(input);
	result = RGBToYIQ(result1);
    }	
    else if (inSpace == "hsv") && (outSpace == "yuv" ){
	result1 = HSVToRGB(input);
	result = RGBToYUV(result1);
    }	

// RGB in
    else if (inSpace == "rgb") && (outSpace == "cmy" ){
	result = RGBToCMY(input);
    }	
    else if (inSpace == "rgb") && (outSpace == "hls" ){
	result = RGBToHLS(input,rWeight,gWeight,bWeight);
    }	
    else if (inSpace == "rgb") && (outSpace == "hsv" ){
	result = RGBToHSV(input);
    }	
    else if (inSpace == "rgb") && (outSpace == "rgb" ){
	result = input;
    }
    else if (inSpace == "rgb") && (outSpace == "yiq" ){
	result = RGBToYIQ(input);
    }	
    else if (inSpace == "rgb") && (outSpace == "yuv" ){
	result = RGBToYUV(input);
    }	


// YIQ in
    else if (inSpace == "yiq") && (outSpace == "cmy" ){
	result1 = YIQToRGB(input);
	result = RGBToCMY(result1);
    }	
    else if (inSpace == "yiq") && (outSpace == "hls" ){
	result1 = YIQToRGB(input);
	result = RGBToHLS(result1);
    }	
    else if (inSpace == "yiq") && (outSpace == "hsv" ){
	result1 = YIQToRGB(input);
	result = RGBToHSV(result1);
    }	
    else if (inSpace == "yiq") && (outSpace == "rgb" ){
	result = YIQToRGB(input);
    }	
    else if (inSpace == "yiq") && (outSpace == "yiq" ){
	result = input;
    }
    else if (inSpace == "yiq") && (outSpace == "yuv" ){
	result1 = YIQToRGB(input);
	result = RGBToYUV(result1);
    }

// YUV in
    else if (inSpace == "yuv") && (outSpace == "cmy" ){
	result1 = YUVToRGB(input);
	result = RGBToCMY(result1);
    }	
    else if (inSpace == "yuv") && (outSpace == "hls" ){
	result1 = YUVToRGB(input);
	result = RGBToHLS(result1);
    }	
    else if (inSpace == "yuv") && (outSpace == "hsv" ){
	result1 = YUVToRGB(input);
	result = RGBToHSV(result1);
    }	
    else if (inSpace == "yuv") && (outSpace == "rgb" ){
	result = YUVToRGB(input);
    }	
    else if (inSpace == "yuv") && (outSpace == "yiq" ){
	result1 = YUVToRGB(input);
	result = RGBToYIQ(result1);
    }
    else if (inSpace == "yuv") && (outSpace == "yuv" ){
	result = input;
    }

    return result;
}

//LogLin function wrapper on Delog/Log functions
//Peter Warner 11/08/00

image LogLin(
	image In=0,
	int conversion=0,
	float rOffset=0,
	float gOffset=0,
	float bOffset=0,
	float rBlack=95,
	float gBlack=rBlack,
	float bBlack=rBlack,
	float rWhite=685,
	float gWhite=rWhite,
	float bWhite=rWhite,
	float rNGamma=0.6,
	float gNGamma=rNGamma,
	float bNGamma=rNGamma,
	float rDGamma=1.7,
	float gDGamma=rDGamma,
	float bDGamma=rDGamma,
	float rSoftClip=0,
	float gSoftClip=rSoftClip,
	float bSoftClip=rSoftClip
)
{
	toto = In;
   if (conversion) {
	    toto = LogRGB(toto, rOffset, gOffset, bOffset, rBlack, gBlack, 
        bBlack, rWhite, gWhite, bWhite, rNGamma, gNGamma, bNGamma, 
        rDGamma, gDGamma, bDGamma, rSoftClip, gSoftClip, bSoftClip);

	} else  {
	    toto = DelogRGB(toto, rOffset, gOffset, bOffset, rBlack, gBlack, 
        bBlack, rWhite, gWhite, bWhite, rNGamma, gNGamma, bNGamma, 
        rDGamma, gDGamma, bDGamma, rSoftClip, gSoftClip, bSoftClip);

    }
    return toto;
}

image BabuCore(
image	In= 0,
int	steps,
int	directionFilter,
float	directionFilterAngle
)
{
    image result= Black(1,1);
    if(directionFilter)
    {
	Rot= Rotate(In,::directionFilterAngle);
	Conv= Convolve(Rot, "rgba", "BabuH", 100, 0);
	result= Rotate(Conv, -::directionFilterAngle);
    }
    else
    {
	for(int i=0;i<steps;++i)
	{
	    Rot= Rotate(In,i*360.0/(float)::steps);
	    Conv= Convolve(Rot, "rgba", "BabuH", 100, 0);
	    Babu= Rotate(Conv, -i*360.0/(float)::steps);
	    result=Max(result,Babu);
	}
    }
    Normalized= Brightness(result,0.1);
    return Normalized;
}

image SobelCore(
image	In= 0,
int	directionFilter,
float	directionFilterAngle
)
{
    AdjustCW= Rotate(In, ::directionFilter ? -::directionFilterAngle: 0);
	SobelV= Convolve(AdjustCW, "rgba", "sobelV", 100, 0);
	SobelH= Convolve(AdjustCW, "rgba", "sobelH", 100, 0);
	SobelV_2= IMult(SobelV, SobelV, 1, 100, 0);
	SobelH_2= IMult(SobelH, SobelH, 1, 100, 0);
	SobelVH_2= IAdd(SobelV_2, SobelH_2, 1, 100);
	Sobel= Select(1+::directionFilter,SobelVH_2,SobelH_2);
    AdjustCCW= Rotate(Sobel, ::directionFilter ? ::directionFilterAngle: 0);
    SquareRoot= Gamma(AdjustCCW,2);
    Normalized= Brightness(SquareRoot,0.25);
    return Normalized;
}

image EdgeDetect(
image	    In= 0,			// Input image
float	    strength= 1,		// Strengthen weaker edges
float	    threshold= 0,		// Only keep edges with strength above threshold
int	    binaryOutput= 0,		// Output 100% white edges
int	    directionFilter= 0,		// Directional edge filter On/Off
float	    directionFilterAngle=0,	// Angle for directional edge filter
int	    despeckle= 0,		// Remove outliers
float	    xBlur= 0,			// Blur level -- X
float	    yBlur= xBlur,		// Blur level -- Y
float	    xDilateErode= 0,		// DilateErode -- X
float	    yDilateErode= xDilateErode,	// DilateErode -- Y
int	    rgbEdges= 0,		// Perform edge extraction on all 3 channels
int	    addEdgesToImage= 0,		// Recombine edges with original image
const char  *method= "Sobel",		// Edge extraction method
int	    babuSteps= 8,		// Number of steps for the babu algorithm
int	    bytes= 1			// Output depth
)
{
    /* Crank up depth and possibly make a mono image
    -------------------------------------------------*/
    UpDepth= Bytes(In, 4);
    Mono= Monochrome(UpDepth,0.30,0.59,0.11);
    MonoOrRGB= Select(1+::rgbEdges,Mono,UpDepth);

    /* Extract image edges
    -----------------------*/
    Babu= BabuCore(MonoOrRGB,::babuSteps,::directionFilter,::directionFilterAngle);
    Sobel= SobelCore(MonoOrRGB,::directionFilter,::directionFilterAngle);
    Edges= Select(1+(::method=="Babu" ? 0 : 1),Babu,Sobel);
    
    /* Massage gradient
    --------------------*/
    Strengthen= Gamma(Edges,::strength);
    CutOff= Threshold(
	Strengthen,
	::threshold,
	::threshold,
	::threshold,
	::threshold,
	::binaryOutput
    );

    /* Outliers, lowpass, morpho
    -----------------------------*/
    for(i=0; i<despeckle; ++i) CutOff= Median(CutOff);
    Blurred= Blur(CutOff,::xBlur,::yBlur,1);
    Thinned= DilateErode(
	Blurred,
	"rgba",
	::xDilateErode,
	::yDilateErode,
	0,
	0
    );

    /* Recombine and return
    ------------------------*/
    DownDepth= Bytes(Thinned,::bytes);
    Reassemble= Screen(In,DownDepth);
    return Select(1+::addEdgesToImage,DownDepth,Reassemble);
}
// Kernels

DefKernel(
    "blur3x3",
    3, 3,
    1/9.0,
    0,
    1, 1, 1,
    1, 1, 1,
    1, 1, 1
);
 
DefKernel(
    "blur5x5",
    5, 5,
    1/25.0,
    0,
    1, 1, 1, 1, 1,
    1, 1, 1, 1, 1,
    1, 1, 1, 1, 1,
    1, 1, 1, 1, 1,
    1, 1, 1, 1, 1
);

DefKernel(
    "sharpen",
    3, 3,
    1,
    0,
    -1, -1, -1,
    -1,  9, -1,
    -1, -1, -1
);
 
DefKernel(
    "edge3x3",
    3, 3,
    1,
    0,
    -1, -1, -1,
    -1,  8, -1,
    -1, -1, -1
);
 
DefKernel(
    "edge5x5",
    5, 5,
    1,
    0,
    -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1,
    -1, -1, 24, -1, -1,
    -1, -1, -1, -1, -1,
    -1, -1, -1, -1, -1
);
 
DefKernel(
    "laplace",
    3, 3,
    1,
    0,
    -sqrt(2)/2, -1, -sqrt(2)/2,
    -1, 4+2*sqrt(2), -1,
    -sqrt(2)/2, -1, -sqrt(2)/2
);
 
DefKernel(
    "smoothedge",
    5, 5,
    1,
    0,
    -0.0625, -0.125, -0.25, -0.125, -0.0625,
    -0.125,  -0.25,  -0.5,  -0.25,  -0.125,
    -0.25,   -0.5,    5.0,  -0.5,   -0.25,
    -0.125,  -0.25,  -0.5,  -0.25,  -0.1251,
    -0.0625, -0.125, -0.25, -0.125, -0.0625
);

DefKernel(
    "sobelH",
    3, 3,
    1,
    0,
    -1,-2,-1,
     0, 0, 0,
     1, 2, 1
);

DefKernel(
    "sobelV",
    3, 3,
    1,
    0,
    -1, 0, 1,
    -2, 0, 2,
    -1, 0, 1
);

DefKernel(
    "BabuV",
    5, 5,
    1,
    0,
    -1,   -1,   0,  1,	1,
    -1,   -1,   0,  1,	1,
    -1,   -1,   0,  1,	1,
    -1,   -1,   0,  1,	1,
    -1,   -1,   0,  1,	1
);

DefKernel(
    "BabuH",
    5, 5,
    1,
    0,
    -1,   -1,  -1,  -1,	 -1,
    -1,   -1,  -1,  -1,	 -1,
     0,    0,   0,   0,   0,
     1,    1,   1,   1,	  1,
     1,    1,   1,   1,	  1
);


// Set this to a value > 0 on slower hardware .

sys.hardwareType = 0;

// Number of columns in the dobox
// This must be set before nrui.h is parsed; so override
// it in startup/.h file, rather than startup/ui.h file
// gui.doBoxColumns = 8;

// these set the default time min / max values.  they need to be in this 
// file because the widget that use them are created before nrui.h is parsed
gui.timeRangeMin = 1;
gui.timeRangeMax = 100;



// this plug sets the amount of memory available to nodes when processing.
// it can be changed from the command line with -mem
 
cache.cacheMemory = 64;

// these plugs control the size and location of the disk cache
// diskCache.cacheMemory controls how much memory is reserved 
// for disk buffering

// diskCache.cacheLocation = "/var/tmp/Shake/cache";	// for IRIX
// diskCache.cacheLocation = "c:/temp/Shake/cache";	// for Windows NT
// diskCache.cacheMaxFile = 2048;
// diskCache.cacheMaxFileSize = 2048*2048*4*2;

diskCache.cacheSize = 512;
diskCache.cacheMemory = 64;

// default frame rate. This affects playback speed and timecode display.
// it can be changed from the command line with -fps

script.framesPerSecond = 24;

// Setting this plug to zero disables automatic UNC file name generation.
// When UNC file names are off, all file names are written exactly as they
// appear in their respective text fields.

script.uncFileNames = 1;


// These defines timecodeModes.
DefTimecodeMode("24 FPS",24);
DefTimecodeMode("25 FPS",25);
DefTimecodeMode("30 FPS DF",30,2,60,600);
DefTimecodeMode("30 FPS ND",30);

// This sets the default format parameters: width, height, aspectRatio, viewerAspectRatio, framesPerSecond,fieldRendering)
// Valid entries includ "Custom" and any user defined format.
DefFormatType("Academy",	    1828, 1556, 1,	    1,	    24,	    "24 FPS");
DefFormatType("CinemaScope",	    1828, 1556, 1/2,	    2,	    24,	    "24 FPS");
DefFormatType("Full",		    2048, 1556, 1,	    1,	    24,	    "24 FPS");
DefFormatType("1.85",		    1828, 1332, 1,	    1,	    24,	    "24 FPS");
DefFormatType("HDTV 1080i/p 30 FPS", 1920, 1080, 1,         1,      30,     "30 FPS ND" );
DefFormatType("HDTV 1080i/p 29.97 FPS ND", 1920, 1080, 1,   1,      29.97,  "30 FPS ND" );
DefFormatType("HDTV 1080i/p 29.97 FPS DF", 1920, 1080, 1,   1,      29.97,  "30 FPS DF" );
DefFormatType("HDTV 1080i/p 25 FPS", 1920, 1080, 1,         1,      25,     "25 FPS" );
DefFormatType("HDTV 1080i/p 24 FPS", 1920, 1080, 1,         1,      24,     "24 FPS" );
DefFormatType("NTSC ND (D1 4:3)",    720,  486, 1/.9f,	    0.9f,   29.97,  "30 FPS ND");
DefFormatType("NTSC DF (D1 4:3)",    720,  486, 1/.9f,	    0.9f,   29.97,  "30 FPS DF");
DefFormatType("NTSC ND (16:9)",	     720,  486, 1/1.2f,	    1.2f,   29.97,  "30 FPS ND");
DefFormatType("NTSC DF (16:9)",	     720,  486, 1/1.2f,	    1.2f,   29.97,  "30 FPS DF");
DefFormatType("PAL (D1 4:3)",	     720,  576, 1/1.066f,   1.066f, 25,	    "25 FPS");
DefFormatType("PAL (16:9)",	     720,  576, 1/1.422f,   1.422f, 25,	    "25 FPS");
DefFormatType("PAL (Square)",	     768,  576, 1,	    1,	    25,	    "25 FPS");

// Change the following values to override some of the individual values in
// a Format setting.  The Format will change to "Custom" if no match is
// found in the defined formats.
// script.defaultWidth= 720;
// script.defaultHeight= 486;
script.timecodeMode =	"24 FPS";
script.format =		"Full";
script.defaultWidth= 720;
script.defaultHeight= 486;
gui.timecodeDisplay = 0;

// Change these to allow top down image writing. This is significantly faster
// but few other applications support it.

script.cineonTopDown = 1;
script.tiffTopDown = 1;

script.jpegQuality= 100;

script.defaultBytes= 1;

//
// To define a curve, four functions must be defined:
//   timeCurveFunc, valueCurveFunc, leftTangentFunc, rightTangentFunc
//   They need to follow the following convention:
//
//   timeCurveFunc	- ${name}(implicit float time, int cycle, const float [numValsPerCV], ....);
//   valueCurveFunc	- ${name}V(float, int cycle, const float [numValsPerCV], ....);
//   leftTangentFunc	- ${name}LTan(const float [2],....);
//   rightTangentFunc	- ${name}RTan(const float [2],....);
//
// The timeCurveFunc and valueCurveFunc functions should encapsulate the 
// same curve value function.  When using the time version within Shake, 
// the first argument is implicitly set to time and should not be entered.
//
// The rightTangentFunc and leftTangentFunc functions return the tangents
// at end conditions given the curve's keys and values.
//
//
// the numValsPerCV needs to be one of the following:
// 2 - key,val
// 4 - key,val,ltan,rtan
//
// the numControlPoint is the number of points needed to compute curve
// vals and tangents.   The default is 3 (current point plus 2 adjacent)
// the max number of control points allowed is 5

DefCurveType("Hermite",4,3);
DefCurveType("Linear", 2,3);
DefCurveType("CSpline",2,3);
DefCurveType("JSpline",2,3);
DefCurveType("NSpline",2,3);
DefCurveType("Step",2,3);


// 
// To add additional mixers to the Transition node, use nfxDefMix(name,source) 
// where source needs to take two image inputs, and provide an output.  It also
// needs an input float parameter name mixPercent to control the amount of mix.
//
image HWipe(image i1=0,image i2=0,float blur=0,int reverse=0,float mixPercent="HermiteV(x,1,[0,50,50]@0,[100,50,50]@100)")
{
    Color1 = Color(max(i1.width,i2.width), max(i1.height,i2.height),1, 1, red, red, 1, 0);
    Crop1 = Crop(Color1, 0, 0, width, height);
     Pan1 = Pan(Crop1, mixPercent*width/100*(reverse?-1:1));
    BlurMe = Blur(Pan1,blur,0,1);
    IMult1 = IMult(i1, BlurMe, 1, 100, 0);
    Invert1 = Invert(BlurMe, "rgba");
    IMult2 = IMult(Invert1, i2, 1, 100, 0);
    IAdd1 = IAdd(IMult1, IMult2, 1, 100);

    return IAdd1;
}

image VWipe(image i1=0,image i2=0,float blur=0,int reverse=0,float mixPercent="HermiteV(x,1,[0,50,50]@0,[100,50,50]@100)")
{
    Color1 = Color(max(i1.width,i2.width),max(i1.height,i2.height),1, 1, red, red, 1, 0);
    Crop1 = Crop(Color1, 0, 0, width, height);
    Pan1 = Pan(Crop1, 0, mixPercent*height/100*(reverse?-1:1));
    BlurMe = Blur(Pan1,0,blur,1);
    IMult1 = IMult(i1, BlurMe, 0, 100, 0);
    Invert1 = Invert(BlurMe, "rgba");
        IMult2 = IMult(Invert1, i2, 1, 100, 0);
    IAdd1 = IAdd(IMult1, IMult2, 1, 100);

    return IAdd1;
}

//nfxDefMixer("inverseDissolve","InverseDissolve()");
nfxDefMixer("horizontalWipe","HWipe()");
nfxDefMixer("verticalWipe","VWipe()");


